"""
_piece.py
The Piece and Source classes.
"""

# ======================================================================

import re

from gspread.utils import (
    a1_range_to_grid_range as gridrange,
    rowcol_to_a1,
)
from loguru import logger

from ._shared import error

# ======================================================================

__all__ = ('Piece',)

# ======================================================================

HYPERLINK_RE = re.compile(r'=HYPERLINK\("(.+)", "(.+)"\)')

# ======================================================================


def _col_str(col):
    """Return the column string of a column number (1-indexed)."""
    return rowcol_to_a1(1, col)[:-1]

# ======================================================================


def _hyperlink(text, link=None):
    """Create a hyperlink formula with optional link.

    Args:
        text (str): The placeholder text.
        link (str): The link.

    Returns:
        str: The hyperlink formula.
    """

    if link is None:
        return text
    escaped = text.replace('"', '\\"')
    return f'=HYPERLINK("{link}", "{escaped}")'


def _parse_hyperlink(hyperlink):
    """Extract the link and text from a hyperlink formula.

    Args:
        hyperlink (str): The hyperlink formula.

    Returns:
        Union[Tuple[str, str], Tuple[None, None]]:
            The link and the text, or None and None if the hyperlink
            is invalid.
    """

    match = HYPERLINK_RE.match(hyperlink)
    if match is None:
        return None, None
    link, text = match.groups()
    return link, text.replace('\\"', '"')


# ======================================================================

def _max(a, b):
    """Find the max of two possibly-None values."""
    if a is None and b is None:
        return None
    if a is None:
        return b
    if b is None:
        return a
    return max(a, b)

# ======================================================================


class Source:
    """Source class."""

    def __init__(self, kwargs):
        """Initialize a source from a JSON dict.

        Args:
            kwargs (Dict): The JSON dict.

        Raises:
            ValueError: If any required key is not found.
            ValueError: If the bar count is not positive.
        """

        for key in ('name', 'link'):
            if key not in kwargs:
                raise ValueError(f'key "{key}" not found')

        self._name = kwargs['name']
        self._link = kwargs['link']
        self._bar_count = kwargs.get('barCount', None)
        # only allow positive bar counts
        if self._bar_count is not None and self._bar_count <= 0:
            raise ValueError('bar count must be positive')
        self._is_supplemental = kwargs.get('supplemental', False)

    @property
    def name(self):
        return self._name

    @property
    def link(self):
        return self._link

    @property
    def bar_count(self):
        return self._bar_count

    @property
    def is_supplemental(self):
        return self._is_supplemental

    def combine(self, other):
        """Combine this source with another by taking max bar count."""
        if self._link != other.link:
            logger.warning('Differing link: "{}"', other.link)
        self._bar_count = _max(self._bar_count, other.bar_count)
        if not self._is_supplemental:
            self._is_supplemental = other.is_supplemental

    def to_json(self, include_supplemental=False):
        """Return a JSON representation of this source."""
        values = {
            'name': self._name,
            'link': self._link,
        }
        if self._bar_count is not None:
            values['barCount'] = self._bar_count
        if include_supplemental and self._is_supplemental:
            values['supplemental'] = True
        return values

    def hyperlink(self):
        """Return a string formula for the linked source."""
        return _hyperlink(self._name, self._link)


class Piece:
    """Piece class."""

    def __init__(self, kwargs, template=None):
        """Initialize a piece from a JSON dict.

        Args:
            kwargs (Dict): The JSON dict.
            template (Optional[Dict]): The template settings.
                If None, only represents a piece; cannot do anything
                with it (such as create a sheet).

        Raises:
            ValueError: If any required key is not found.
            ValueError: If any source's bar count is not positive.
        """

        for key in ('title', 'sources'):
            if key not in kwargs:
                raise ValueError(f'key "{key}" not found')

        self._name = kwargs['title']
        self._link = kwargs.get('link', None)
        self._initial_bar_count = kwargs.get('barCount', None)

        sources = []
        for i, args in enumerate(kwargs['sources']):
            try:
                source = Source(args)
            except ValueError as ex:
                # re-raise the exception with added text
                ex.args = (f'source {i}: ' + ex.args[0],) + ex.args[1]
                raise
            sources.append(source)

        self._only_supplemental = False
        self._sources = {}
        self._supplemental_sources = {}
        self._bar_count = None
        self._add_sources(sources)

        self._template = template

        if self._template is None:
            self._values = None
            return

        # row 1 in `create_sheet()`

        before_bars = []
        # all other rows from template
        before_bars += [
            [header]
            for header in template['metaDataFields'].values()
        ]
        # empty row
        before_bars.append([])

        # bars section goes here

        after_bars = [
            # empty row
            [],
            # comments row
            [template['commentFields']['comments']]
        ]

        # before bar section, after bar section
        self._values = [before_bars, after_bars]

    @property
    def name(self):
        return self._name

    @property
    def link(self):
        return self._link

    @property
    def sources(self):
        return list(self._sources.values())

    @property
    def final_bar_count(self):
        if self._bar_count is None:
            return self._template['values']['defaultBarCount']
        return self._bar_count

    @property
    def only_supplemental(self):
        return self._only_supplemental

    def _add_sources(self, sources):
        """Add sources."""

        # add all sources
        for source in sources:
            name = source.name
            if name in self._sources:
                logger.debug(
                    'Combining source "{}" in piece "{}"',
                    name, self._name
                )
                existing_source = self._sources[name]
                existing_source.combine(source)
                if existing_source.is_supplemental:
                    self._supplemental_sources[name] = existing_source
                    self._sources.pop(name)
                continue
            if name in self._supplemental_sources:
                logger.debug(
                    'Combining source "{}" in piece "{}"',
                    name, self._name
                )
                self._supplemental_sources[name].combine(source)
                continue
            if source.is_supplemental:
                self._supplemental_sources[name] = source
            else:
                self._sources[name] = source

        # ignore this piece if there are only supplemental sources
        self._only_supplemental = (
            len(self._sources) == 0 and
            len(self._supplemental_sources) > 0
        )

        # re-calculate bar count with non-supplemental sources
        self._bar_count = self._initial_bar_count
        for source in self._sources.values():
            self._bar_count = _max(self._bar_count, source.bar_count)

    def combine(self, other):
        """Combine this piece with another by combining all sources."""
        if self._link is None:
            self._link = other.link
        elif other.link is None:
            logger.warning('Missing link')
        elif self._link != other.link:
            logger.warning('Differing link: "{}"', other.link)

        if self._initial_bar_count is None:
            # pylint: disable=protected-access
            self._initial_bar_count = other._initial_bar_count

        self._add_sources(other.sources)

    def has_source(self, name):
        """Check if this piece has a source with the given name."""
        return name in self._sources

    def get_source(self, name):
        """Get the source by the given name, or None."""
        return self._sources.get(name, None)

    def to_json(self, include_supplemental=False):
        """Return a JSON representation of this piece."""
        values = {}
        values['title'] = self._name
        if self._link is not None:
            values['link'] = self._link
        if self._initial_bar_count is not None:
            values['barCount'] = self.final_bar_count
        values['sources'] = [
            source.to_json()
            for source in self._sources.values()
        ]
        if include_supplemental:
            values['sources'] += [
                source.to_json(include_supplemental=True)
                for source in self._supplemental_sources.values()
            ]
        return values

    def _create_supplemental_sources_column(self, values, is_master):
        """Create the supplemental sources column(s) if needed.
        Wraps to more columns if there are too many sources.
        Return True if created.
        """
        if len(self._supplemental_sources) == 0:
            return False

        num_cols = len(values[0])
        start_row = 2 if is_master else 1
        # don't go into the comments row
        wrap_row = len(values) - 1

        values[0].append(
            self._template['commentFields']['supplementalSources']
        )

        row_index = start_row
        for source in self._supplemental_sources.values():
            row = values[row_index]
            if len(row) < num_cols:
                row += [''] * (num_cols - len(row))
            row.append(source.hyperlink())

            row_index += 1
            if row_index == wrap_row:
                # wrap back to top
                row_index = start_row

        return True

    def create_sheet(self, spreadsheet):
        """Create a sheet for this piece.

        Args:
            spreadsheet (gspread.Spreadsheet): The parent spreadsheet.

        Returns:
            gspread.Worksheet: The created sheet.
        """

        if self._template is None:
            raise RuntimeError(
                f'cannot create sheet for piece "{self._name}": '
                'no template given during creation'
            )

        # add sheet
        sheet = spreadsheet.add_worksheet(self._name, 1, 1)

        # complete row 1
        row1 = (
            [_hyperlink(self._name, self._link)] +
            [source.hyperlink() for source in self._sources.values()] +
            [self._template['commentFields']['notes']]
        )

        # proper bar count
        bar_count = self.final_bar_count

        values = [
            row1,
            # use copy of `self._values`
            *[row[:] for row in self._values[0]],
            *[[i + 1] for i in range(bar_count)],
            *[row[:] for row in self._values[1]],
        ]

        notes_col = len(values[0])
        blank_row1 = 1 + len(self._values[0])  # row 1 + headers
        blank_row2 = blank_row1 + bar_count + 1
        comments_row = blank_row2 + 1

        # to make sure columns are auto-resized with a minimum width,
        # insert text in every column, then remove it after formatting
        resize = self._template['volunteerSpreadsheet']['resize']
        if resize:
            values[blank_row1 - 1] = (
                # header column (202 pixels)
                ['placeholderplaceholderplaceh'] +
                # source columns (154 pixels)
                ['placeholderplaceholde'] * (len(values[0]) - 2)
            )

        has_supplemental_col = self._create_supplemental_sources_column(
            values, is_master=False
        )

        # put the values
        sheet.update(values, raw=False)

        _format_sheet(
            spreadsheet, sheet, self._template,
            notes_col, blank_row1, blank_row2, comments_row,
            resize=resize, has_supplemental_col=has_supplemental_col
        )

        if resize:
            start = f'A{blank_row1}'
            end = f'{_col_str(notes_col - 1)}{blank_row1}'
            sheet.batch_clear([f'{start}:{end}'])

        return sheet

    def create_master_sheet(self, spreadsheet, piece_data):
        """Create a sheet for the master spreadsheet.

        Args:
            spreadsheet (gspread.Spreadsheet): The master spreadsheet.
            piece_data (Dict): The data for the piece.

        Returns:
            gspread.Worksheet: The created sheet.
        """

        if self._template is None:
            raise RuntimeError(
                f'cannot create sheet for piece "{self._name}": '
                'no template given during creation'
            )

        # add sheet
        sheet = spreadsheet.add_worksheet(self._name, 1, 1)

        # proper bar count
        bar_count = self.final_bar_count

        # finish headers
        values = [
            [_hyperlink(self._name, self._link)],
            ['Volunteer'],
            # use copy of `self._values`
            *[row[:] for row in self._values[0]],
            *[[i + 1] for i in range(bar_count)],
            *[row[:] for row in self._values[1]],
        ]

        headers = tuple(self._template['metaDataFields'].keys())

        headers_start = 2
        headers_iter = [
            (headers_start + i, header)
            for i, header in enumerate(headers)
        ]

        blank_row1 = 2 + len(headers) + 1  # 2 rows + headers
        blank_row2 = blank_row1 + bar_count + 1
        comments_row = blank_row2 + 1

        bars_start = blank_row1
        bars_iter = [
            (bars_start + i, str(i + 1))
            for i in range(bar_count)
        ]

        # go through sources and add
        sources = piece_data['sources']
        # list of starting col number for each source (1-indexed)
        source_cols = []
        col = 2
        for name, source in self._sources.items():
            source_data = sources[name]
            source_cols.append(col)

            def add_column(email, volunteer):
                # row 1
                values[0].append('')
                # row 2
                values[1].append(email)
                # headers
                for row, header in headers_iter:
                    values[row].append(volunteer[header])
                # bars
                bars = volunteer['bars']
                for row, bar_num in bars_iter:
                    values[row].append(bars[bar_num])
                # comments
                values[comments_row - 1].append(volunteer['comments'])

            start_col = col - 1

            # volunteers
            for email, volunteer in source_data['volunteers'].items():
                add_column(email, volunteer)
                col += 1
            # summary
            add_column(
                self._template['commentFields']['summary'],
                source_data['summary']
            )
            col += 1

            # put the source hyperlink
            values[0][start_col] = source.hyperlink()

        # notes column
        notes_col = col

        def note_str(notes):
            return '\n'.join(
                f'{email}: {note}'
                for email, note in notes.items()
            )

        values[0].append(self._template['commentFields']['notes'])
        notes = piece_data['notes']
        for row, header in headers_iter:
            values[row].append(note_str(notes[header]))
        bars = notes['bars']
        for row, bar_num in bars_iter:
            values[row].append(note_str(bars[bar_num]))

        # force headers column to be 202 pixels minimum
        resize = self._template['masterSpreadsheet']['resize']
        if resize:
            values[blank_row1 - 1] = ['placeholderplaceholderplaceh']

        has_supplemental_col = self._create_supplemental_sources_column(
            values, is_master=True
        )

        # put the values
        sheet.update(values, raw=False)

        _format_sheet(
            spreadsheet, sheet, self._template,
            notes_col, blank_row1, blank_row2, comments_row,
            resize=resize, is_master=True, source_cols=source_cols,
            has_supplemental_col=has_supplemental_col
        )

        if resize:
            sheet.batch_clear([f'A{blank_row1}'])

        return sheet

    @staticmethod
    def export_sheet(sheet, template):
        """Export piece data from a sheet.
        Assumes same format as a created sheet from
        `Piece.create_sheet()`.
        Ignores the supplemental sources column if it exists.

        Args:
            sheet (gspread.Worksheet): The sheet.
            template (Dict): The template settings.

        Returns:
            Tuple[bool, Dict]: Whether the export was successful,
                and the data in JSON format.
        """
        ERROR_RETURN = False, None

        def _error(msg, *args, **kwargs):
            return error(msg.format(*args, **kwargs), ERROR_RETURN)

        try:
            (
                values,
                piece_link, piece_name,
                headers_iter, bars_iter,
                comments_row, notes_col,
            ) = _export_helper(sheet, template, is_master=False)
        except Exception as e:  # pylint: disable=broad-except
            return _error(
                'Error while exporting sheet "{}": {}', sheet.title, e
            )

        def export_column(col, include_comments=True):
            column = {}
            for row, header in headers_iter:
                column[header] = values[row][col]
            bars = {}
            for row in bars_iter:
                bars[values[row][0]] = values[row][col]
            column['bars'] = bars
            if include_comments:
                column['comments'] = values[comments_row][col]
            return column

        sources = []
        for col in range(1, notes_col):
            link, name = _parse_hyperlink(values[0][col])
            if link is None:
                return _error(
                    'sheet "{}": column {} doesn\'t have a valid '
                    'hyperlink',
                    sheet.title, _col_str(col + 1)
                )

            source = {
                'name': name,
                'link': link,
                **export_column(col),
            }
            sources.append(source)

        notes = export_column(notes_col, False)

        return True, {
            'title': piece_name,
            'link': piece_link,
            'sources': sources,
            'notes': notes,
        }

    @staticmethod
    def export_master_sheet(sheet, template):
        """Export a sheet from the master spreadsheet.
        Assumes same format as a created sheet from
        `Piece.create_master_sheet()`.
        Assumes the last column for each source is the summary column.

        Args:
            sheet (gspread.Worksheet): The sheet.
            template (Dict): The template settings.

        Returns:
            Tuple[bool, Dict]: Whether the export was successful,
                and the data in JSON format.
        """
        ERROR_RETURN = False, None

        def _error(msg, *args, **kwargs):
            return error(msg.format(*args, **kwargs), ERROR_RETURN)

        try:
            (
                values,
                piece_link, piece_name,
                headers_iter, bars_iter,
                comments_row, notes_col,
            ) = _export_helper(sheet, template, is_master=True)
        except Exception as e:  # pylint: disable=broad-except
            return _error(
                'Error while exporting master sheet "{}": {}',
                sheet.title, e
            )

        sources = []
        curr_source = {}
        for col in range(1, notes_col):
            if values[0][col] != '':
                # new source
                link, name = _parse_hyperlink(values[0][col])
                if link is None:
                    return _error(
                        'sheet "{}": column {} doesn\'t have a valid '
                        'hyperlink',
                        sheet.title, _col_str(col + 1)
                    )

                curr_source = {
                    'name': name,
                    'link': link,
                    'volunteers': {},
                    'summary': None,
                }
                sources.append(curr_source)

            # always overwrite summary, pushing existing to volunteers
            email = values[1][col]

            column = {}
            for row, header in headers_iter:
                column[header] = values[row][col]
            bars = {}
            for row in bars_iter:
                bars[values[row][0]] = values[row][col]
            column['bars'] = bars
            column['comments'] = values[comments_row][col]

            if curr_source['summary'] is not None:
                # push to volunteers
                prev_email, prev_column = curr_source['summary']
                curr_source['volunteers'][prev_email] = prev_column
            # save to summary
            curr_source['summary'] = (email, column)

        # fix all summaries
        for source in sources:
            _, summary = source['summary']
            source['summary'] = summary

        def parse_note(note_str):
            note = {}
            for line in note_str.split('\n'):
                if line == '':
                    continue
                try:
                    email, text = line.split(': ', 1)
                    note[email] = text
                except ValueError:
                    logger.warning(
                        'Note line has invalid format: {}', line
                    )
            return note

        notes = {}
        for row, header in headers_iter:
            notes[header] = parse_note(values[row][notes_col])
        bars = {}
        for row in bars_iter:
            bars[values[row][0]] = parse_note(values[row][notes_col])
        notes['bars'] = bars

        return True, {
            'title': piece_name,
            'link': piece_link,
            'sources': sources,
            'notes': notes,
        }

# ======================================================================


def _format_sheet(spreadsheet, sheet, template,
                  notes_col, blank_row1, blank_row2, comments_row,
                  resize=False, is_master=False, source_cols=None,
                  has_supplemental_col=False
                  ):
    """Format a piece sheet."""

    sheet_id = sheet.id

    def hex_to_rgb(hex_color):
        """Changes a hex color code (no pound) to an RGB color dict.
        Each value is a fraction decimal instead of an integer.
        """
        colors = ('red', 'green', 'blue')
        return {
            key: int(hex_color[i:i+2], 16) / 255
            for key, i in zip(colors, range(0, 6, 2))
        }

    BLACK = hex_to_rgb('000000')

    header_end = 2 if is_master else 1

    if source_cols is None:
        source_cols = []
    source_cols.append(notes_col)

    requests = []

    # make everything except last (comments) row middle aligned
    requests.append({
        'repeatCell': {
            'range': {
                'sheetId': sheet_id,
                'endRowIndex': blank_row2 - 1,
            },
            'cell': {
                'userEnteredFormat': {
                    'verticalAlignment': 'MIDDLE',
                },
            },
            'fields': 'userEnteredFormat.verticalAlignment',
        }
    })

    # formatting
    bolded = {
        'cell': {
            'userEnteredFormat': {
                'textFormat': {'bold': True},
            },
        },
        'fields': 'userEnteredFormat.textFormat.bold',
    }
    centered = {
        'cell': {
            'userEnteredFormat': {
                'horizontalAlignment': 'CENTER',
            },
        },
        'fields': 'userEnteredFormat.horizontalAlignment',
    }
    middle_bolded = {
        'cell': {
            'userEnteredFormat': {
                'verticalAlignment': 'MIDDLE',
                'textFormat': {'bold': True},
            },
        },
        'fields': ','.join((
            'userEnteredFormat.verticalAlignment',
            'userEnteredFormat.textFormat.bold',
        ))
    }
    wrapped_top_align = {
        'cell': {
            'userEnteredFormat': {
                'wrapStrategy': 'WRAP',
                'verticalAlignment': 'TOP',
            },
        },
        'fields': ','.join((
            'userEnteredFormat.wrapStrategy',
            'userEnteredFormat.verticalAlignment',
        ))
    }
    source_end_column = _col_str(notes_col - 1)
    range_formats = [
        # first row (piece name, sources, notes, possible supplemental)
        ('A1:1', bolded),
        # headers
        (f'A2:A{blank_row1 - 1}', bolded),
        # sources
        (f'B1:{source_end_column}1', centered),
        # comments header
        (f'A{comments_row}', middle_bolded),
        # comments row
        (f'B{comments_row}:{comments_row}', wrapped_top_align),
    ]
    if is_master:
        # make all summary cells be centered
        center_align = {
            'cell': {
                'userEnteredFormat': {
                    'horizontalAlignment': 'CENTER',
                },
            },
            'fields': 'userEnteredFormat.horizontalAlignment',
        }
        range_formats += [
            (f'{_col_str(col - 1)}2', center_align)
            for col in source_cols[1:]
        ]
        # make volunteers line bolded
        range_formats += [
            (f'B2:{source_end_column}2', bolded),
        ]
    for range_name, fmt in range_formats:
        requests.append({
            'repeatCell': {
                'range': gridrange(range_name, sheet_id=sheet_id),
                **fmt,
            }
        })

    # borders
    range_borders = []

    if is_master:
        double_border = {
            'style': 'DOUBLE',
            'color': BLACK,
        }

        # double border after the header rows
        range_borders += [
            (f'{header_end}:{header_end}', {'bottom': double_border}),
        ]

        # double border before every source column
        left_double_border = {'left': double_border}
        for col in source_cols:
            col_str = _col_str(col)
            range_borders.append(
                (f'{col_str}:{col_str}', left_double_border)
            )

        # merge source columns
        for i in range(len(source_cols) - 1):
            requests.append({
                'mergeCells': {
                    'range': {
                        'sheetId': sheet_id,
                        # row 1
                        'startRowIndex': 0,
                        'endRowIndex': 1,
                        # between each source col (0-indexed)
                        'startColumnIndex': source_cols[i] - 1,
                        'endColumnIndex': source_cols[i + 1] - 1,
                    },
                    'mergeType': 'MERGE_ALL',
                }
            })

    # borders around empty rows
    solid_black = {
        'style': 'SOLID',
        'color': BLACK,
    }
    top_bottom_border = {
        'top': solid_black,
        'bottom': solid_black,
    }
    range_borders += [
        (f'A{blank_row1}:{blank_row1}', top_bottom_border),
        (f'A{blank_row2}:{blank_row2}', top_bottom_border),
    ]

    # dotted border after every fifth bar
    interval = 5
    bottom_dotted_border = {
        'bottom': {
            'style': 'DOTTED',
            'color': BLACK,
        }
    }
    range_borders += [
        (f'A{row}:{row}', bottom_dotted_border) for row in
        range(blank_row1 + interval, blank_row2 - 1, interval)
    ]

    for range_name, borders in range_borders:
        requests.append({
            'updateBorders': {
                'range': gridrange(range_name, sheet_id=sheet_id),
                **borders,
            }
        })

    column_widths = []
    if not resize:
        column_widths.append(
            # column 1: width 200
            ({'startIndex': 0, 'endIndex': 1}, 200),
        )
    if is_master or not resize:
        column_widths.append(
            # source columns: width 150
            ({'startIndex': 1, 'endIndex': notes_col - 1}, 150),
        )
    column_widths.append(
        # notes col: width 300
        ({'startIndex': notes_col - 1, 'endIndex': notes_col}, 300),
    )
    if has_supplemental_col:
        column_widths.append(
            # supplemental sources column(s): width of 150
            ({'startIndex': notes_col}, 150),
        )
    for pos, width in column_widths:
        requests.append({
            'updateDimensionProperties': {
                'properties': {
                    'pixelSize': width,
                },
                'fields': 'pixelSize',
                'range': {
                    'sheetId': sheet_id,
                    'dimension': 'COLUMNS',
                    **pos
                },
            }
        })

    if resize:
        end_col = 1 if is_master else (notes_col - 1)
        # auto-resize headers column and source columns
        requests.append({
            'autoResizeDimensions': {
                'dimensions': {
                    'sheetId': sheet_id,
                    'dimension': 'COLUMNS',
                    'startIndex': 0,
                    'endIndex': end_col,
                },
            }
        })

    # make comments row proper height
    requests.append({
        'updateDimensionProperties': {
            'properties': {
                'pixelSize': template['values']['commentsRowHeight'],
            },
            'fields': 'pixelSize',
            'range': {
                'sheetId': sheet_id,
                'dimension': 'ROWS',
                'startIndex': comments_row - 1,  # 0-indexed here
                'endIndex': comments_row,
            },
        }
    })

    # freeze header rows and column 1
    requests.append({
        'updateSheetProperties': {
            'properties': {
                'sheetId': sheet_id,
                'gridProperties': {
                    'frozenRowCount': header_end,
                    'frozenColumnCount': 1,
                },
            },
            'fields': ','.join((
                'gridProperties.frozenRowCount',
                'gridProperties.frozenColumnCount',
            )),
        }
    })

    # default banding style white/gray
    white_gray_banding = {
        'rowProperties': {
            'headerColor': hex_to_rgb('bdbdbd'),
            'firstBandColor': hex_to_rgb('ffffff'),
            'secondBandColor': hex_to_rgb('f3f3f3'),
            # 'footerColor': hex_to_rgb('dedede'),
        },
    }
    # don't include last column (notes) or last row (comments)
    requests.append({
        'addBanding': {
            'bandedRange': {
                'range': {
                    'sheetId': sheet_id,
                    'startRowIndex': 0,
                    'endRowIndex': comments_row - 1,
                    'startColumnIndex': 1,
                    'endColumnIndex': notes_col - 1,
                },
                **white_gray_banding,
            },
        }
    })

    # data validation
    header_rows = {
        header: header_end + i
        for i, header in enumerate(template['metaDataFields'].keys())
    }
    for key, validation in template['validation'].items():
        row = header_rows[key]
        v_type = validation['type']
        if v_type == 'checkbox':
            condition = {'type': 'BOOLEAN'}
        elif v_type == 'dropdown':
            condition = {
                'type': 'ONE_OF_LIST',
                'values': [
                    {'userEnteredValue': val}
                    for val in validation['values']
                ],
            }
        else:
            # shouldn't happen
            continue
        requests.append({
            'setDataValidation': {
                'range': {
                    'sheetId': sheet_id,
                    'startRowIndex': row,
                    'endRowIndex': row + 1,
                    'startColumnIndex': 1,
                    'endColumnIndex': notes_col - 1,
                },
                'rule': {
                    'condition': condition,
                    'strict': False,
                    'showCustomUi': True,
                },
            }
        })

    body = {'requests': requests}
    spreadsheet.batch_update(body)

# ======================================================================


def _export_helper(sheet, template, is_master):
    """Export a sheet and get important helper values."""

    values = sheet.get_values(value_render_option='formula')

    # row 1
    row1 = values[0]

    piece_link, piece_name = _parse_hyperlink(row1[0])
    if piece_name is None:
        piece_name = row1[0]

    headers = tuple(template['metaDataFields'].keys())

    start_row = 2 if is_master else 1
    headers_range = (start_row, start_row + len(headers))
    headers_iter = list(zip(range(*headers_range), headers))

    bars_range = (start_row + len(headers) + 1, len(values) - 2)
    bars_iter = list(range(*bars_range))

    comments_row = len(values) - 1

    # skip over supplemental sources column
    notes_col = len(row1) - 1
    notes_col_title = template['commentFields']['notes']
    while notes_col >= 0 and row1[notes_col] != notes_col_title:
        notes_col -= 1
    if notes_col == -1:
        # not found, so assume it's the last column
        notes_col = len(row1) - 1

    return (
        values,
        piece_link, piece_name,
        headers_iter, bars_iter,
        comments_row, notes_col,
    )



"""
export_master.py
Export the master spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._shared import error
from ._input_files import (
    read_template,
)
from ._output_files import (
    read_spreadsheets_index,
    write_summary,
)
from ._sheets import (
    gspread_auth,
    open_spreadsheet,
)
from ._piece import Piece

# ======================================================================

__all__ = ('export_master',)

# ======================================================================

MASTER_KEY = 'MASTER'

# ======================================================================


@click.command(
    'export_master',
    help='Export the master spreadsheet.'
)
@click.option(
    '-si', type=str,
    help='A filepath to replace the spreadsheets index file.'
)
@click.option(
    '-td', type=str,
    help='A filepath to replace the template definitions file.'
)
@click.option(
    '-s', 'summary_path', type=str,
    help='A filepath to replace the summary file.'
)
@click.option(
    '--strict', is_flag=True, default=False, flag_value=True,
    help='Fail on warnings instead of only displaying them.'
)
def export_master(si, td, summary_path, strict):
    """Export the master spreadsheet.

    Args:
        si (str): A filepath to replace the spreadsheets index file.
        td (str): A filepath to replace the template definitions file.
        summary_path (str): A filepath to replace the summary file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    success, gc = gspread_auth()
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return

    if MASTER_KEY not in spreadsheets:
        error(
            f'Master spreadsheet (key "{MASTER_KEY}") not found in '
            'spreadsheets index file'
        )
        return

    success, template = read_template(td, strict)
    if not success:
        return

    success, spreadsheet = open_spreadsheet(
        gc, spreadsheets[MASTER_KEY]
    )
    if not success:
        return

    summary = []
    for sheet in spreadsheet.worksheets():
        success, piece_data = Piece.export_master_sheet(
            sheet, template
        )
        if not success:
            return
        summary.append(piece_data)

    success = write_summary(summary, summary_path)
    if not success:
        return

    logger.info('Done')



"""
_input_files.py
Handles input files.
"""

# ======================================================================

from loguru import logger

from ._shared import fail_on_warning, error
from ._read_write import (
    _read_default,
    read_json,
    write_json,
    read_settings,
)
from ._piece import Piece
from ._piece_data import PieceData
from ._volunteer import Volunteer

# ======================================================================

__all__ = (
    'read_template',
    'read_piece_definitions', 'fix_piece_definitions',
    'read_volunteer_definitions', 'fix_volunteer_definitions',
)

# ======================================================================

# the options for the "publicAccess" field
PUBLIC_ACCESS_OPTIONS = (None, 'view', 'edit')


def read_template(path=None, strict=False):
    """Read the template definitions file.

    Args:
        path (Optional[str]): A path to the template definitions file to
            use instead.
            Default is None (use the settings file).
        strict (bool): Whether to fail on warnings instead of only
            displaying them.

    Returns:
        Tuple[bool, Dict]: Whether the read was successful,
            and the template settings.
    """
    ERROR_RETURN = False, None

    def _error(msg):
        return error(msg, ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info('Reading template definitions file')

    key = 'template'

    try:
        raw_values = read_json(key, path)
    except (FileNotFoundError, ValueError) as ex:
        return _error(ex)
    defaults = _read_default('template_definitions.json')
    values = {
        level: {**level_defaults}
        for level, level_defaults in defaults.items()
    }
    invalid = False
    warning = False

    # no required fields

    # get values from file
    for level, level_values in raw_values.items():
        if level == 'validation':
            continue
        if level not in values:
            warning = True
            logger.warning('unknown key "{}"', level)
            continue
        level_defaults = values[level]
        for k, value in level_values.items():
            if k not in level_defaults:
                warning = True
                logger.warning('unknown key "{}"."{}"', level, k)
                continue
            level_defaults[k] = value
    # validation
    values['validation'] = {}
    for k, validation in raw_values.get('validation', {}).items():
        if k not in values['metaDataFields']:
            warning = True
            logger.warning('"validation": unknown key "{}"', k)
            continue
        if 'type' not in validation:
            warning = True
            logger.warning('"validation"."{}": no type', k)
            continue
        v_type = validation['type']
        if v_type == 'checkbox':
            values['validation'][k] = {'type': 'checkbox'}
        elif v_type == 'dropdown':
            if len(validation.get('values', [])) == 0:
                warning = True
                logger.warning(
                    '"validation"."{}": no values for dropdown', k
                )
                continue
            values['validation'][k] = {
                'type': 'dropdown',
                'values': validation['values'],
            }
        else:
            warning = True
            logger.warning(
                '"validation"."{}"."type": unknown type "{}"', k, v_type
            )

    # validate values
    # public access options
    for ss in ('masterSpreadsheet', 'volunteerSpreadsheet'):
        value = values[ss]['publicAccess']
        if value not in PUBLIC_ACCESS_OPTIONS:
            warning = True
            logger.warning(
                '"{}"."publicAccess": invalid value "{}" '
                '(must be null, "view", or "edit")',
                ss, value
            )
            # reset to default
            values[ss]['publicAccess'] = defaults[ss]['publicAccess']
    # volunteer spreadsheet title must have "{email}" at most once
    ss = 'volunteerSpreadsheet'
    if values[ss]['title'].count('{email}') > 1:
        invalid = True
        _error(
            f'"{ss}"."title": can only contain "{{email}}" at most once'
        )
    # must be a boolean
    for k in ('shareWithVolunteer', 'resize'):
        val = values[ss][k]
        if val not in (True, False):
            warning = True
            logger.warning(
                '"{}"."{}": invalid value "{}" (must be true or false)',
                ss, k, val
            )
            # reset to default
            values[ss][k] = defaults[ss][k]
    # default bar count must be positive
    if values['values']['defaultBarCount'] <= 0:
        invalid = True
        _error('"values"."defaultBarCount": must be positive')
    # comments row height must be at least 21
    if values['values']['commentsRowHeight'] < 21:
        invalid = True
        _error('"values"."commentsRowHeight": must be at least 21')

    if invalid:
        return ERROR_RETURN
    if strict and warning:
        fail_on_warning()
        return ERROR_RETURN

    return True, values

# ======================================================================


def _piece_definitions(path, template, msg):
    """Read the piece definitions file.
    Returns whether successful, whether there were pieces with no
    sources, the pieces with at least one source, and the pieces that
    only have supplemental sources.
    """
    ERROR_RETURN = False, None, None, None

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info('{} piece definitions file', msg)

    try:
        raw_pieces = read_json('pieces', path)
    except (FileNotFoundError, ValueError) as ex:
        return _error(ex)
    if len(raw_pieces) == 0:
        return _error('no pieces found')

    pieces = {}
    for i, args in enumerate(raw_pieces):
        try:
            piece = Piece(args, template)
        except ValueError as ex:
            return _error('piece {}: {}', i, ex)

        name = piece.name
        if name in pieces:
            # combine sources with previous piece
            logger.debug('Combining piece "{}"', name)
            pieces[name].combine(piece)
        else:
            pieces[name] = piece

    filtered = {}
    only_supplemental = {}
    no_sources = False
    for title, piece in pieces.items():
        if piece.only_supplemental:
            logger.warning(
                'piece "{}": only supplemental sources found', title
            )
            only_supplemental[title] = piece
            continue
        if len(piece.sources) == 0:
            no_sources = True
            _error('piece "{}": no sources found', title)
            continue
        filtered[title] = piece

    return True, no_sources, filtered, only_supplemental


def read_piece_definitions(template,
                           path=None,
                           as_data=False,
                           as_both=False
                           ):
    """Read the piece definitions file.
    Repeated pieces and sources will be combined.
    All supplemental sources will be ignored, including pieces with only
    supplemental sources.

    Args:
        template (Dict): The template settings.
        path (Optional[str]): A path to the piece definitions file to
            use instead.
            Default is None (use the settings file).
        as_data (bool): Whether to return the piece objects as data.
            Default is False.
        as_both (bool): Whether to return the piece objects as both
            regular objects and as data.
            If True, returns a tuple of piece objects and piece data in
            place of the mapping.
            Default is False.

    Returns:
        Tuple[bool, Dict[str, Union[Piece, PieceData]]]:
            Whether the read was successful,
            and a mapping from piece names to piece objects.
    """

    success, no_sources, pieces, _ = _piece_definitions(
        path, template, msg='Reading'
    )
    if not success or no_sources:
        return False, None

    if as_both or as_data:
        pieces_data = {
            title: PieceData(piece)
            for title, piece in pieces.items()
        }
        if as_both:
            return True, (pieces, pieces_data)
        pieces = pieces_data

    return True, pieces


def fix_piece_definitions(path=None):
    """Fix the piece definitions file.
    Repeated pieces and sources will be combined.
    Supplemental sources will be moved to the end.
    Pieces with only supplemental sources will be moved to the end.

    Args:
        path (Optional[str]): A path to the piece definitions file to
            use instead.
            Default is None (use the settings file).

    Returns:
        Tuple[bool, Dict[str, Piece]]: Whether the fix was successful,
            and a mapping from piece names to piece objects.
    """
    ERROR_RETURN = False, None

    success, _, pieces, only_supplemental = _piece_definitions(
        path, None, msg='Fixing'
    )
    if not success:
        return ERROR_RETURN

    # add supplemental pieces to end
    for title, piece in only_supplemental.items():
        pieces[title] = piece

    if len(pieces) == 0:
        error('No valid pieces to write back to piece definitions file')
        return ERROR_RETURN

    data = [
        piece.to_json(include_supplemental=True)
        for piece in pieces.values()
    ]
    write_json(
        'pieces', data, path,
        msg='fixed piece definitions'
    )

    return True, pieces

# ======================================================================


def _volunteer_definitions(pieces, path, msg):
    """Read the volunteer definitions file.
    Ignores pieces that only had supplemental sources in the piece
    definitions file.
    Returns whether successful, whether there were volunteers with no
    pieces, whether there were unknown pieces, and the volunteers with
    at least one piece.
    """
    ERROR_RETURN = False, None, None, None

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info('{} volunteer definitions file', msg)

    try:
        raw_volunteers = read_json('volunteers', path)
    except (FileNotFoundError, ValueError) as ex:
        return _error(ex)
    if len(raw_volunteers) == 0:
        return _error('no volunteers found')

    volunteers = {}
    unknown_pieces = False
    for i, args in enumerate(raw_volunteers):
        try:
            volunteer = Volunteer(args, pieces)
        except ValueError as ex:
            return _error('volunteer {}: {}', i, ex)

        email = volunteer.email
        if email in volunteers:
            # combine pieces with previous volunteer
            logger.debug('Combining volunteer "{}"', email)
            volunteers[email].combine(volunteer)
        else:
            volunteers[email] = volunteer

        for piece in volunteer.unknown_pieces:
            unknown_pieces = True
            logger.warning(
                'volunteer "{}": unknown piece "{}"', email, piece
            )

    filtered = {}
    no_pieces = False
    for email, volunteer in volunteers.items():
        if len(volunteer.pieces) == 0:
            no_pieces = True
            _error('volunteer "{}": no pieces found', email)
            continue
        filtered[email] = volunteer

    return True, no_pieces, unknown_pieces, filtered


def read_volunteer_definitions(pieces, path=None, strict=False):
    """Read the volunteer definitions file.
    Repeated volunteer emails will be combined.
    Unknown pieces and pieces with only supplemental sources will be
    ignored.

    Args:
        pieces (Dict[str, Piece]): The known pieces.
        path (Optional[str]): A path to the volunteer definitions file
            to use instead.
            Default is None (use the settings file).

    Returns:
        Tuple[bool, Dict[str, Volunteer]]:
            Whether the read was successful,
            and a mapping from volunteer emails to volunteer objects.
    """
    ERROR_RETURN = False, None

    success, no_pieces, unknown_pieces, volunteers = \
        _volunteer_definitions(pieces, path, msg='Reading')
    if not success or no_pieces:
        return ERROR_RETURN
    if strict and unknown_pieces:
        fail_on_warning()
        return ERROR_RETURN

    return True, volunteers


def fix_volunteer_definitions(pieces, path=None):
    """Fix the volunteer definitions file.
    Repeated volunteer emails will be combined.
    Pieces with only supplemental sources will be removed.

    Args:
        pieces (Dict[str, Piece]): The known pieces.
        path (Optional[str]): A path to the volunteer definitions file
            to use instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the fix was successful.
    """
    ERROR_RETURN = False

    success, _, _, volunteers = _volunteer_definitions(
        pieces, path, msg='Fixing'
    )
    if not success:
        return ERROR_RETURN

    if len(volunteers) == 0:
        error(
            'No valid volunteers to write back to volunteer '
            'definitions file'
        )
        return ERROR_RETURN

    data = [
        volunteer.to_json()
        for volunteer in volunteers.values()
    ]
    write_json(
        'volunteers', data, path,
        msg='fixed volunteer definitions'
    )

    return True



"""
compile_pieces.py
Compile all piece JSON data flies into a single file.
"""

# ======================================================================

import click
from loguru import logger

from ._shared import error
from ._input_files import (
    read_template,
    read_piece_definitions,
)
from ._output_files import (
    read_piece_data,
    write_summary,
)

# ======================================================================

__all__ = ('compile_pieces',)

# ======================================================================


@click.command(
    'compile_pieces',
    help='Compile all piece JSON data flies into a single file.'
)
@click.option(
    '-td', type=str,
    help='A filepath to replace the template definitions file.'
)
@click.option(
    '-pd', type=str,
    help='A filepath to replace the piece definitions file.'
)
@click.option(
    '-pdp', type=str,
    help=(
        'A filepath to replace the piece data path file. '
        'Must include "{piece}".'
    )
)
@click.option(
    '-s', 'summary_path', type=str,
    help='A filepath to replace the summary file.'
)
@click.option(
    '--strict', is_flag=True, default=False, flag_value=True,
    help='Fail on warnings instead of only displaying them.'
)
def compile_pieces(td, pd, pdp, summary_path, strict):
    """Compile all piece JSON data flies into a single file.

    Args:
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        pdp (str): A filepath to replace the piece data path file.
            Must include "{piece}" exactly once.
        summary_path (str): A filepath to replace the summary file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    # validate args
    if pdp is not None and pdp.count('{piece}') != 1:
        error('`pdp` must include "{piece}" exactly once')
        return

    logger.warning(
        'For most accurate summary, run the `piece_summary` command '
        'first.'
    )

    success, template = read_template(td, strict)
    if not success:
        return

    success, pieces = read_piece_definitions(template, pd, as_data=True)
    if not success:
        return

    success, pieces_data = read_piece_data(pieces, pdp, strict)
    if not success:
        return

    summary = []
    for title, piece in pieces_data.items():
        piece_obj = pieces[title]
        # turn sources into list
        piece['sources'] = list(piece['sources'].values())
        for source in piece['sources']:
            source['summary'] = piece_obj.make_default()
        summary.append(piece)

    success = write_summary(summary, summary_path)
    if not success:
        return

    logger.info('Done')



"""
__main__.py
The command-line interface for Roseingrave.
"""

# ======================================================================

import click

from .reauth import reauth
from .fix_input import fix_input
from .create_sheet import create_sheet
from .volunteer_summary import volunteer_summary
from .piece_summary import piece_summary
from .compile_pieces import compile_pieces
from .import_master import import_master
from .export_master import export_master

# ======================================================================


class OrderedGroup(click.Group):
    """Lists the commands in the listed order (instead of alphabetical).
    """

    def __init__(self, *args, **kwargs):
        commands = kwargs.get('commands', None)
        if commands is not None:
            if isinstance(commands, list):
                kwargs['commands'] = {cmd.name: cmd for cmd in commands}
        super().__init__(*args, **kwargs)

    def list_commands(self, ctx):
        return list(self.commands.keys())


cli = OrderedGroup(
    commands=[
        reauth,
        fix_input,
        create_sheet,
        volunteer_summary,
        piece_summary,
        compile_pieces,
        import_master,
        export_master,
    ],
    help='Massively scalable musical source comparator'
)

if __name__ == '__main__':
    cli()



"""
_output_files.py
Handles output files.
"""
# Logs all unexpected exceptions
# pylint: disable=broad-except

# ======================================================================

import re
from itertools import zip_longest
from pathlib import Path

from loguru import logger

from ._shared import fail_on_warning, error
from ._read_write import (
    get_path,
    read_json,
    write_json,
    write_json_file,
    read_settings,
)
from ._sheets import (
    gspread_auth,
    open_spreadsheet,
)

# ======================================================================

__all__ = (
    'read_spreadsheets_index', 'write_spreadsheets_index',
    'fix_spreadsheets_index',
    'read_volunteer_data', 'write_volunteer_data',
    'read_piece_data', 'write_piece_data',
    'read_summary', 'write_summary',
)

# ======================================================================


def read_spreadsheets_index(path=None):
    """Read the spreadsheets index file.

    Args:
        path (Optional[str]): A path to the spreadsheets index file to
            use instead.
            Default is None (use the settings file).

    Returns:
        Tuple[bool, Dict[str, str]]: Whether the read was successful,
            and a mapping from volunteer emails to spreadsheet links,
            or an empty dict if the file does not exist.
    """
    ERROR_RETURN = False, None

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info('Reading spreadsheets index file')

    try:
        contents = read_json('spreadsheetsIndex', path)
    except FileNotFoundError:
        contents = {}
    except ValueError as ex:
        error(ex)
        return ERROR_RETURN
    return True, contents


def write_spreadsheets_index(data, path=None):
    """Write to the spreadsheets index file.
    Updates the file if it exists, or creates it if it doesn't exist.

    Args:
        data (Dict[str, str]): The mapping from volunteer emails to
            spreadsheet links.
        path (Optional[str]): A path to the spreadsheets index file to
            use instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the write was successful.
    """

    success = read_settings()
    if not success:
        return False

    write_json(
        'spreadsheetsIndex', data, path,
        msg='created spreadsheet links'
    )

    return True


def fix_spreadsheets_index(path=None):
    """Fix the spreadsheets index file.

    Args:
        path (Optional[str]): A path to the spreadsheets index file to
            use instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the fix was successful.
    """
    ERROR_RETURN = False

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    success = read_settings()
    if not success:
        return False

    logger.info('Fixing spreadsheets index file')

    try:
        contents = read_json('spreadsheetsIndex', path)
    except (FileNotFoundError, ValueError) as ex:
        return _error(ex)

    # check all links
    fixed = {}
    success, gc = gspread_auth()
    if not success:
        return ERROR_RETURN
    for email, link in contents.items():
        success, _ = open_spreadsheet(gc, link)
        if not success:
            if email == 'MASTER':
                volunteer = 'master spreadsheet'
            else:
                volunteer = f'volunteer "{email}"'
            _error('Removing {}', volunteer)
            continue
        fixed[email] = link

    data = {}
    if 'MASTER' in fixed:
        data['MASTER'] = fixed.pop('MASTER')
    for email, link in sorted(fixed.items()):
        data[email] = link

    write_json(
        'spreadsheetsIndex', data, path,
        msg='fixed spreadsheet links'
    )

    return True


# ======================================================================


def _read_format_files(fmt_path, arg):
    """Read all files matching a formatted path.

    Args:
        fmt_path (Path): The format path.
        arg (str): The arg in the path to match.

    Raises:
        ValueError: If `fmt_path` doesn't include `arg` exactly once.

    Returns:
        Tuple[bool, Dict[str, Union[Dict, List]]]:
            Whether there were any files found, and a mapping from each
            file's arg to the file's parsed contents.
    """
    fmt_path = fmt_path.resolve()

    if str(fmt_path).count(arg) != 1:
        raise ValueError(
            f'path "{fmt_path}" must include "{arg}" exactly once'
        )

    path_parts = fmt_path.parts
    cwd_parts = Path('.').resolve().parts

    # find where they start to differ and where the arg occurs
    differ = None
    arg_index = None
    path_re = None
    for i, (path_part, part) in \
            enumerate(zip_longest(path_parts, cwd_parts)):
        if path_part is not None and arg in path_part:
            arg_index = i
            path_re = re.compile(path_part.replace(arg, '(.+)'))
            if differ is None:
                differ = i
            break

        if differ is None:
            if path_part is None or part is None or path_part != part:
                differ = i

        if differ is not None and arg_index is not None:
            break

    # arg from filepath -> file contents
    data = {}

    num_parts = len(path_parts)

    def check_path(index, curr_path):
        """Check if the current path matches the format path so far,
        and populates data as it goes.
        """

        last_part = index >= num_parts - 1

        if (
            # invalid index
            index >= num_parts or
            # can only be a file when last part
            curr_path.is_file() != last_part
        ):
            return

        current = curr_path.name
        found_arg = None

        if index == arg_index:
            # check match
            match = path_re.match(current)
            if match is None:
                return
            found_arg = match.groups()[0]
        elif current != path_parts[index]:
            return

        # last part
        if last_part:
            if found_arg is not None:
                # add to data
                data[found_arg] = read_json('', str(curr_path))
            return

        # recursively check next
        for path in curr_path.iterdir():
            check_path(index + 1, path)

    # search all files
    for path in Path(*path_parts[:differ]).iterdir():
        check_path(differ, path)

    if len(data) == 0:
        logger.info('No matching files found')
        return False, None

    return True, data

# ======================================================================


def _validate_source(piece,
                     sources,
                     index,
                     raw_source,
                     p_loc,
                     strict,
                     has_volunteers=False,
                     has_summary=False
                     ):
    """Validate a source from an input file.
    Adds the fixed source to `sources`.

    Args:
        piece (PieceData): The piece the source belongs to.
        sources (Dict[str, Dict]): The target sources.
        index (int): The index of this source in the sources array.
        raw_source (Dict): The source JSON dict.
        p_loc (str): The piece location, for error messages.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
        has_volunteers (bool): Whether the raw source has volunteer
            information.
            If False, `raw_source` is expected to have all required
            column fields.
            Default is False.
        has_summary (bool): Whether the raw source has summary
            information.
            If True, `raw_source` is expected to have a "summary" field.
            Default is False.

    Returns:
        bool: Whether there was an error.
    """
    SUCCESS_RETURN = False
    SKIP_RETURN = False
    ERROR_RETURN = True

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    fields = ('name', 'link')
    if has_volunteers:
        fields += ('volunteers',)
    if has_summary:
        fields += ('summary',)
    missing_fields = [
        key for key in fields
        if key not in raw_source
    ]
    if len(missing_fields) > 0:
        return _error(
            'Missing fields {} for {}, source {}',
            ','.join(f'"{key}"' for key in missing_fields),
            p_loc, index
        )

    name = raw_source['name']
    s_link = raw_source['link']

    if not piece.has_source(name):
        logger.warning(
            'Unknown source "{}" for {} '
            '(not in piece definitions file)',
            name, p_loc
        )
        if strict:
            fail_on_warning()
            return ERROR_RETURN
        return SKIP_RETURN

    if name in sources:
        logger.warning('Repeated source "{}" for {}', name, p_loc)
        if strict:
            fail_on_warning()
            return ERROR_RETURN
        return SKIP_RETURN

    s_loc = f'{p_loc}, source "{name}"'

    if s_link != piece.get_source(name).link:
        logger.warning(
            'Incorrect source link "{}" for {}', s_link, s_loc
        )
        if strict:
            fail_on_warning()
            return ERROR_RETURN

    if has_volunteers:
        volunteers = {}
        for email, volunteer in raw_source['volunteers'].items():
            v_loc = f'{s_loc}, volunteer "{email}"'
            warning, fixed = piece.with_defaults(volunteer, v_loc)
            if strict and warning:
                fail_on_warning()
                return ERROR_RETURN
            volunteers[email] = fixed
        adding = {'volunteers': volunteers}
    else:
        warning, source = piece.with_defaults(raw_source, s_loc)
        if strict and warning:
            fail_on_warning()
            return ERROR_RETURN
        adding = source

    if has_summary:
        warning, summary = piece.with_defaults(
            raw_source['summary'], s_loc
        )
        if strict and warning:
            fail_on_warning()
            return ERROR_RETURN
        adding['summary'] = summary

    sources[name] = {
        'name': name,
        'link': s_link,
        **adding,
    }

    return SUCCESS_RETURN


def _validate_piece(pieces,
                    fixed_pieces,
                    v_loc,
                    raw_piece,
                    strict,
                    index=None,
                    from_piece_file=False,
                    file=None,
                    has_summary=False
                    ):
    """Validate a piece from an input file.
    Adds the fixed piece to `fixed_pieces`.

    Args:
        pieces (Dict[str, PieceData]): The known pieces.
        fixed_pieces (Dict[str, Dict]): The target pieces.
        v_loc (Optional[str]): The location of this piece's volunteer,
            for error messages.
            If None, volunteer location is not included in messages.
        raw_piece (Dict): The piece JSON dict.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
        index (Optional[str[]): The index of this piece.
            Default is None.
        from_piece_file (bool): Whether the input is from a piece file.
            Default is False.
        file (Optional[str]): The name of the file the input came from.
            Default is None.
        has_summary (bool): Whether the sources have summary
            information.
            Default is False.

    Returns:
        bool: Whether there was an error.
    """
    SUCCESS_RETURN = False
    SKIP_RETURN = False
    ERROR_RETURN = True

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    missing_fields = [
        key for key in ('title', 'link', 'sources', 'notes')
        if key not in raw_piece
    ]
    if len(missing_fields) > 0:
        if from_piece_file and file is not None:
            loc = f'piece file "{file}"'
        elif index is not None:
            loc = f'piece {index}'
            if v_loc is not None:
                loc = f'{v_loc}, {loc}'
        else:
            loc = 'piece (location unknown)'
        return _error(
            'Missing fields {} for {}',
            ','.join(f'"{key}"' for key in missing_fields), loc
        )

    title = raw_piece['title']
    p_link = raw_piece['link']

    if from_piece_file and file is not None and file != title:
        msg = f'Piece name "{title}" doesn\'t match file arg "{file}"'
        if strict:
            logger.warning(msg)
            fail_on_warning()
            return ERROR_RETURN
        logger.warning('{}; Using "{}" for piece', msg, title)

    if title not in pieces:
        msg = f'Unknown piece "{title}"'
        if v_loc is not None:
            msg += f' for {v_loc}'
        msg += ' (not in piece definitions file)'
        logger.warning(msg)
        if strict:
            fail_on_warning()
            return ERROR_RETURN
        return SKIP_RETURN
    piece_obj = pieces[title]

    if title in fixed_pieces:
        msg = f'Repeated piece "{title}"'
        if v_loc is not None:
            msg += f' for {v_loc}'
        logger.warning(msg)
        if strict:
            fail_on_warning()
            return ERROR_RETURN
        return SKIP_RETURN

    p_loc = f'piece "{title}"'
    if v_loc is not None:
        p_loc = f'{v_loc}, {p_loc}'

    warning = False
    if piece_obj.link is None:
        if p_link is not None:
            warning = True
            logger.warning(
                'Extra piece link "{}" for {}', p_link, p_loc
            )
    else:
        if p_link is None:
            warning = True
            logger.warning('Missing piece link for {}', p_loc)
        elif p_link != piece_obj.link:
            warning = True
            logger.warning(
                'Incorrect piece link "{}" for {}', p_link, p_loc
            )
    if strict and warning:
        fail_on_warning()
        return ERROR_RETURN

    sources = {}
    for i, raw_source in enumerate(raw_piece['sources']):
        had_error = _validate_source(
            piece_obj, sources, i, raw_source, p_loc, strict,
            has_volunteers=from_piece_file, has_summary=has_summary
        )
        if had_error:
            return ERROR_RETURN
    missing_sources = [
        name for name in piece_obj.all_sources()
        if name not in sources
    ]
    if len(missing_sources) > 0:
        return _error(
            'Missing sources {} for {}',
            ','.join(f'"{name}"' for name in missing_sources), p_loc
        )

    warning, notes = piece_obj.with_defaults(
        raw_piece['notes'], p_loc,
        exclude_comments=True, is_notes=from_piece_file
    )
    if strict and warning:
        fail_on_warning()
        return ERROR_RETURN

    fixed_pieces[title] = {
        'title': title,
        'link': p_link,
        'sources': sources,
        'notes': notes,
    }

    return SUCCESS_RETURN

# ======================================================================


def read_volunteer_data(pieces, fmt_path=None, strict=False):
    """Read volunteer data from files.
    Unknown and repeated pieces and sources will be ignored.

    Args:
        pieces (Dict[str, PieceData]): A mapping from piece names to
            piece data.
        fmt_path (Optional[str]): A format path for volunteer data to
            use instead.
            Default is None (use the settings file).
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.

    Returns:
        Tuple[bool, Dict[str, List]]: Whether the read was successful,
            and a mapping from volunteer emails to data.
    """
    ERROR_RETURN = False, None

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    # validate args
    try:
        path = get_path(
            'volunteerDataPath', fmt_path,
            must_exist=False
        )
    except Exception as ex:
        return _error(ex)
    if str(path).count('{email}') != 1:
        return _error(
            'Path to volunteer data files must include "{email}" '
            'exactly once'
        )

    logger.info('Reading volunteer data from files: {}', path)

    success, raw_data = _read_format_files(path, '{email}')
    if not success:
        return ERROR_RETURN

    # validate data
    volunteers = {}
    for email, volunteer_pieces in raw_data.items():
        v_loc = f'volunteer "{email}"'
        fixed_pieces = {}
        for i, raw_piece in enumerate(volunteer_pieces):
            had_error = _validate_piece(
                pieces, fixed_pieces, v_loc, raw_piece, strict, i
            )
            if had_error:
                return ERROR_RETURN
        volunteers[email] = fixed_pieces

    return True, volunteers


def write_volunteer_data(data, fmt_path=None):
    """Write volunteer data to files.
    Replaces files if they already exist.

    Args:
        data (Dict[str, List]): A mapping from volunteer emails to data.
        fmt_path (Optional[str]): A format path for volunteer data to
            use instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the write was successful.
    """

    def _error(msg):
        return error(msg, False)

    success = read_settings()
    if not success:
        return False

    # validate args
    try:
        path = str(get_path(
            'volunteerDataPath', fmt_path,
            must_exist=False
        ))
    except Exception as ex:
        return _error(ex)
    if path.count('{email}') != 1:
        return _error(
            'Path to volunteer data files must include "{email}" '
            'exactly once'
        )

    logger.info('Writing volunteer data to files: {}', path)

    for email, volunteer_data in data.items():
        write_json_file(path.format(email=email), volunteer_data)

    return True

# ======================================================================


def read_piece_data(pieces, fmt_path=None, strict=False):
    """Read piece data from files.

    Args:
        pieces (Dict[str, PieceData]): A mapping from piece names to
            piece data.
        fmt_path (Optional[str]): A format path for piece data to use
            instead.
            Default is None (use the settings file).
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.

    Returns:
        Tuple[bool, Dict[str, Dict]]: Whether the read was successful,
            and a mapping from piece names to data.
    """
    ERROR_RETURN = False, None

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    # validate args
    try:
        path = get_path('pieceDataPath', fmt_path, must_exist=False)
    except Exception as ex:
        return _error(ex)
    if str(path).count('{piece}') != 1:
        return _error(
            'Path to piece data files must include "{piece}" exactly '
            'once'
        )

    logger.info('Reading piece data from files: {}', path)

    success, raw_data = _read_format_files(path, '{piece}')
    if not success:
        return ERROR_RETURN

    # validate data
    # FUTURE: can validate that certain volunteer emails show up?
    #   i.e., all sources must have the same volunteers, and the
    #   emails in "notes" must be a subset
    pieces_data = {}
    for file, raw_piece in raw_data.items():
        had_error = _validate_piece(
            pieces, pieces_data, None, raw_piece, strict,
            from_piece_file=True, file=file
        )
        if had_error:
            return ERROR_RETURN

    return True, pieces_data


def write_piece_data(data, fmt_path=None):
    """Write piece data to files.
    Replaces files if they already exist.

    Args:
        data (Dict[str, Dict]): A mapping from piece names to data.
        fmt_path (Optional[str]): A format path for piece data to use
            instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the write was successful.
    """

    def _error(msg):
        return error(msg, False)

    success = read_settings()
    if not success:
        return False

    # validate args
    try:
        path = str(get_path(
            'pieceDataPath',
            fmt_path, must_exist=False
        ))
    except Exception as ex:
        return _error(ex)
    if path.count('{piece}') != 1:
        return _error(
            'Path to piece data files must include "{piece}" exactly '
            'once'
        )

    logger.info('Writing piece data to files: {}', path)

    for piece, piece_data in data.items():
        write_json_file(path.format(piece=piece), piece_data)

    return True

# ======================================================================


def read_summary(pieces, path=None, strict=False):
    """Read from the summary file.
    Validates the data in the file.

    Args:
        pieces (Dict[str, PieceData]): A mapping from piece names to
            piece data.
        path (Optional[str]): A path to the summary file to use instead.
            Default is None (use the settings file).
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.

    Returns:
        Tuple[bool, Dict[str, Dict]]: Whether the read was successful,
            and a mapping from piece names to piece summary data.
    """
    ERROR_RETURN = False, None

    def _error(msg):
        return error(msg, ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info('Reading summary file')

    key = 'summary'

    try:
        values = read_json(key, path)
    except FileNotFoundError as ex:
        return _error(ex)

    # validate data
    summary = {}
    for i, raw_piece in enumerate(values):
        had_error = _validate_piece(
            pieces, summary, None, raw_piece, strict,
            index=i, from_piece_file=True, has_summary=True
        )
        if had_error:
            return ERROR_RETURN

    return True, summary


def write_summary(summary, path=None):
    """Write to the summary file.

    Args:
        summary (List[Dict]): The summary.
        path (Optional[str]): A path to the summary file to use instead.
            Default is None (use the settings file).

    Returns:
        bool: Whether the write was successful.
    """

    success = read_settings()
    if not success:
        return False

    write_json('summary', summary, path, msg='summary')
    return True



"""
volunteer_summary.py
Export volunteer JSON data files.
"""

# ======================================================================

import click
from loguru import logger

from ._shared import fail_on_warning, error
from ._input_files import (
    read_template,
)
from ._output_files import (
    read_spreadsheets_index,
    write_volunteer_data,
)
from ._sheets import (
    gspread_auth,
    open_spreadsheet,
)
from ._piece import Piece

# ======================================================================

__all__ = ('volunteer_summary',)

# ======================================================================


def export_spreadsheet(gc, email, link, template):
    """Export a volunteer's spreadsheet.

    Args:
        gc (gspread.Client): The client.
        email (str): The volunteer email.
        link (str): The spreadsheet link.
        template (Dict): The template settings.

    Returns:
        Tuple[bool, List[Dict]]: Whether the export was successful,
            and the exported data in JSON format.
    """
    ERROR_RETURN = False, None

    logger.debug('Working on volunteer "{}"', email)

    success, spreadsheet = open_spreadsheet(gc, link)
    if not success:
        return ERROR_RETURN

    data = []
    for sheet in spreadsheet.worksheets():
        success, piece_data = Piece.export_sheet(sheet, template)
        if not success:
            continue
        data.append(piece_data)

    return True, data

# ======================================================================


@click.command(
    'volunteer_summary',
    help='Export volunteer JSON data files.',
)
@click.argument('emails', type=str, nargs=-1)
@click.option(
    '-si', type=str,
    help='A filepath to replace the spreadsheets index file.'
)
@click.option(
    '-td', type=str,
    help='A filepath to replace the template definitions file.'
)
@click.option(
    '-vdp', type=str,
    help=(
        'A filepath to replace the volunteer data path file. '
        'Must include "{email}".'
    )
)
@click.option(
    '--strict', is_flag=True, default=False, flag_value=True,
    help='Fail on warnings instead of only displaying them.'
)
def volunteer_summary(emails, si, td, vdp, strict):
    """Export volunteer JSON data files.

    Args:
        emails (Tuple[str, ...]): The volunteers to export data for. If
            none given, exports data for all volunteers.
        si (str): A filepath to replace the spreadsheets index file.
        td (str): A filepath to replace the template definitions file.
        vdp (str): A filepath to replace the volunteer data path file.
            Must include "{email}" exactly once.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    # validate args
    if vdp is not None and vdp.count('{email}') != 1:
        error('`vdp` must include "{email}" exactly once')
        return

    success, gc = gspread_auth()
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return
    # remove master spreadsheet
    spreadsheets.pop('MASTER', None)

    success, template = read_template(td, strict)
    if not success:
        return

    if len(emails) > 0:
        # filter volunteers in `emails`
        filtered = {}
        for email in emails:
            if email not in spreadsheets:
                logger.warning(
                    'Volunteer "{}" not found in spreadsheets index '
                    'file',
                    email
                )
                continue
            filtered[email] = spreadsheets[email]
        spreadsheets = filtered

    if len(spreadsheets) == 0:
        logger.info('No volunteers to export data for')
        return

    # volunteer email -> data
    data = {}

    # export spreadsheets
    logger.info('Exporting data from spreadsheets')
    for email, link in spreadsheets.items():
        success, volunteer_data = export_spreadsheet(
            gc, email, link, template
        )
        if strict and not success:
            fail_on_warning()
            return
        if success:
            data[email] = volunteer_data

    if len(data) > 0:
        success = write_volunteer_data(data, vdp)
        if not success:
            return

    logger.info('Done')



"""
import_master.py
Update the master spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import (
    read_template,
    read_piece_definitions,
)
from ._output_files import (
    read_summary,
    read_spreadsheets_index, write_spreadsheets_index,
)
from ._sheets import (
    gspread_auth,
    create_spreadsheet,
    open_spreadsheet,
    add_temp_sheet,
    add_permissions,
)

# ======================================================================

__all__ = ('import_master',)

# ======================================================================

MASTER_KEY = 'MASTER'

# ======================================================================


def _create_piece_sheets(spreadsheet, pieces, summary):
    """Create the piece sheets in the master spreadsheet.

    Args:
        spreadsheet (gspread.Spreadsheet): The master spreadsheet.
        pieces (Dict[str, Piece]): A mapping from piece names to piece
            objects.
        summary (Dict[str, Dict]): The piece summary.

    Returns:
        bool: Whether the creations were successful.
    """

    # delete existing sheets
    existing_sheets = list(spreadsheet.worksheets())
    invalid_names = (
        set(summary.keys()) |
        set(sheet.title for sheet in existing_sheets)
    )
    success, temp_sheet = add_temp_sheet(spreadsheet, invalid_names)
    if not success:
        return False
    for sheet in existing_sheets:
        spreadsheet.del_worksheet(sheet)

    for title in summary.keys():
        logger.debug('Creating sheet for piece "{}"', title)
        pieces[title].create_master_sheet(spreadsheet, summary[title])

    # delete temp sheet
    spreadsheet.del_worksheet(temp_sheet)

    return True

# ======================================================================


@click.command(
    'import_master',
    help='Update the master spreadsheet.'
)
@click.option(
    '-c', '--create', is_flag=True, default=False, flag_value=True,
    help='Create a new master spreadsheet.'
)
@click.option(
    '-td', type=str,
    help='A filepath to replace the template definitions file.'
)
@click.option(
    '-pd', type=str,
    help='A filepath to replace the piece definitions file.'
)
@click.option(
    '-s', 'summary_path', type=str,
    help='A filepath to replace the summary file.'
)
@click.option(
    '-si', type=str,
    help='A filepath to replace the spreadsheets index file.'
)
@click.option(
    '--strict', is_flag=True, default=False, flag_value=True,
    help='Fail on warnings instead of only displaying them.'
)
def import_master(create, td, pd, summary_path, si, strict):
    """Update the master spreadsheet.

    Args:
        create (str): Whether to create a new master spreadsheet.
            Default is False.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        summary_path (str): A filepath to replace the summary file.
        si (str): A filepath to replace the spreadsheets index file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    logger.warning(
        'For most accurate summary, run the `compile_pieces` or '
        '`export_master` command first.'
    )

    success, gc = gspread_auth()
    if not success:
        return

    success, template = read_template(td, strict)
    if not success:
        return

    success, (pieces, pieces_data) = read_piece_definitions(
        template, pd, as_both=True
    )
    if not success:
        return

    success, summary = read_summary(pieces_data, summary_path, strict)
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return

    # check if need to create spreadsheet
    if MASTER_KEY not in spreadsheets:
        create = True

    # create or open spreadsheet
    if create:
        ss_settings = template['masterSpreadsheet']

        success, spreadsheet = create_spreadsheet(
            gc, ss_settings['title'], ss_settings['folder']
        )
        if not success:
            return

        # permissions
        success = add_permissions(
            spreadsheet,
            ss_settings['publicAccess'],
            ss_settings['shareWith']
        )
        if not success:
            gc.del_spreadsheet(spreadsheet.id)
            return

        # save link
        spreadsheets[MASTER_KEY] = spreadsheet.url
    else:
        link = spreadsheets[MASTER_KEY]
        success, spreadsheet = open_spreadsheet(gc, link)
        if not success:
            return

    # create pieces
    success = _create_piece_sheets(spreadsheet, pieces, summary)
    if create and not success:
        # delete created spreadsheet
        gc.del_spreadsheet(spreadsheet.id)
        return

    if create:
        success = write_spreadsheets_index(spreadsheets, si)
        if not success:
            return

    logger.info('Done')



"""
create_sheet.py
Create a volunteer spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._shared import fail_on_warning
from ._input_files import (
    read_template,
    read_piece_definitions,
    read_volunteer_definitions,
)
from ._output_files import (
    read_spreadsheets_index, write_spreadsheets_index,
)
from ._sheets import (
    gspread_auth,
    create_spreadsheet,
    open_spreadsheet,
    add_temp_sheet,
    share_spreadsheet,
    add_permissions,
)

# ======================================================================

__all__ = ('create_sheet',)

# ======================================================================


def create_spreadsheets(gc, spreadsheets, template, emails):
    """Create spreadsheets for the given volunteers.

    Args:
        gc (gspread.Client): The client.
        spreadsheets (Dict[str, str]): The existing spreadsheets.
        template (Dict): The template settings.
        emails (Iterable[str]): The emails of the volunteers.

    Returns:
        bool: Whether the creations were successful.
    """

    if len(emails) == 0:
        return True

    logger.info('Creating new spreadsheets')

    created = []

    def delete_created():
        for ss in created:
            gc.del_spreadsheet(ss)

    ss_settings = template['volunteerSpreadsheet']
    folder = ss_settings['folder']
    title_fmt = ss_settings['title']
    public_access = ss_settings['publicAccess']
    share_with_volunteer = ss_settings['shareWithVolunteer']
    share_with = ss_settings['shareWith']

    for email in emails:
        success, spreadsheet = create_spreadsheet(
            gc, title_fmt.format(email=email), folder
        )
        if not success:
            delete_created()
            return False
        created.append(spreadsheet.id)

        success = add_permissions(
            spreadsheet, public_access, share_with
        )
        if not success:
            delete_created()
            return False

        if share_with_volunteer:
            # make volunteer an editor
            msg = (
                'This is an invitation for you to collaborate on a '
                'crowd-sourcing spreadsheet for a project run by '
                'Roseingrave. Thank you for accepting our invitation '
                'as a volunteer.'
            )
            success = share_spreadsheet(
                spreadsheet, email, 'edit', notify=True, msg=msg
            )
            if not success:
                delete_created()
                return False

        spreadsheets[email] = spreadsheet.url

    return True

# ======================================================================


def populate_spreadsheets(gc, spreadsheets, volunteers, pieces, strict):
    """Populate the spreadsheets with the volunteer pieces.

    Args:
        gc (gspread.Client): The client.
        spreadsheets (Dict[str, str]): The spreadsheets.
        volunteers (Dict[str, Volunteer]): The volunteers.
        pieces (Dict[str, Piece]): The pieces.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.

    Returns:
        bool: Whether the population was successful.
    """

    logger.info('Populating spreadsheets with pieces')

    # piece name -> sheet object
    sheets = {}

    for email, volunteer in volunteers.items():
        logger.debug('Working on volunteer "{}"', email)

        link = spreadsheets[email]
        success, spreadsheet = open_spreadsheet(gc, link)
        if not success:
            if strict:
                fail_on_warning()
                return False
            continue

        # delete existing sheets
        existing_sheets = list(spreadsheet.worksheets())
        invalid_names = (
            set(volunteer.pieces) |
            set(sheet.title for sheet in existing_sheets)
        )
        success, temp_sheet = add_temp_sheet(spreadsheet, invalid_names)
        if not success:
            if strict:
                fail_on_warning()
                return False
            continue
        for sheet in existing_sheets:
            spreadsheet.del_worksheet(sheet)

        for piece_name in volunteer.pieces:
            if piece_name not in sheets:
                logger.debug(
                    'Creating sheet for piece "{}"', piece_name
                )
                sheet = pieces[piece_name].create_sheet(spreadsheet)
                sheets[piece_name] = sheet
            else:
                # copy from previously created
                logger.debug('Copying sheet for piece "{}"', piece_name)
                data = sheets[piece_name].copy_to(spreadsheet.id)
                # update sheet title to piece name
                sheet = spreadsheet.get_worksheet_by_id(data['sheetId'])
                sheet.update_title(piece_name)

        # delete temp sheet
        spreadsheet.del_worksheet(temp_sheet)

    return True

# ======================================================================


@click.command(
    'create_sheet',
    help='Create volunteer spreadsheets.',
)
@click.argument('emails', type=str, nargs=-1)
@click.option(
    '-r', '--replace', is_flag=True, default=False, flag_value=True,
    help='Replace existing volunteer spreadsheets.'
)
@click.option(
    '-n', '--new', is_flag=True, default=False, flag_value=True,
    help='Create new spreadsheets for all volunteers.'
)
@click.option(
    '-td', type=str,
    help='A filepath to replace the template definitions file.'
)
@click.option(
    '-pd', type=str,
    help='A filepath to replace the piece definitions file.'
)
@click.option(
    '-vd', type=str,
    help='A filepath to replace the volunteer definitions file.'
)
@click.option(
    '-si', type=str,
    help='A filepath to replace the spreadsheets index file.'
)
@click.option(
    '--strict', is_flag=True, default=False, flag_value=True,
    help='Fail on warnings instead of only displaying them.'
)
def create_sheet(emails, replace, new, td, pd, vd, si, strict):
    """Create volunteer spreadsheets.

    Args:
        emails (Tuple[str, ...]): The volunteers to create spreadsheets
            for. If none given, creates spreadsheets for all volunteers.
        replace (bool): Whether to replace existing volunteer
            spreadsheets.
            Default is False.
        new (bool): Whether to create new spreadsheets for all
            volunteers.
            Default is False.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        vd (str): A filepath to replace the volunteer definitions file.
        si (str): A filepath to replace the spreadsheets index file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    success, gc = gspread_auth()
    if not success:
        return

    success, template = read_template(td, strict)
    if not success:
        return

    success, pieces = read_piece_definitions(template, pd)
    if not success:
        return

    success, volunteers = read_volunteer_definitions(pieces, vd, strict)
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return

    if len(emails) > 0:
        # filter volunteers in `emails`
        filtered = {}
        for email in emails:
            # don't allow this to update the master spreadsheet
            if email == 'MASTER':
                continue
            if email not in volunteers:
                logger.warning(
                    'Volunteer "{}" not found in volunteer definitions '
                    'file',
                    email
                )
                continue
            filtered[email] = volunteers[email]
        volunteers = filtered

    if new:
        # create new spreadsheets for all volunteers,
        # even if they already have a spreadsheet
        create_for = set(volunteers.keys())
    else:
        # don't create new for emails that already have a spreadsheet
        already_exist = {
            email for email in volunteers
            if email in spreadsheets
        }
        create_for = set(volunteers.keys()) - already_exist

        # if `replace` is false, don't edit existing spreadsheets
        if not replace:
            for email in already_exist:
                logger.debug(
                    'Volunteer "{}" being skipped '
                    '(already in spreadsheets index file)',
                    email
                )
                volunteers.pop(email)

    if len(volunteers) == 0:
        logger.info('No volunteers to create spreadsheets for')
        return

    # create new spreadsheets
    success = create_spreadsheets(
        gc, spreadsheets, template, create_for
    )
    if not success:
        return

    # populate spreadsheets
    success = populate_spreadsheets(
        gc, spreadsheets, volunteers, pieces, strict
    )
    if not success:
        return

    if len(create_for) > 0:
        success = write_spreadsheets_index(spreadsheets, si)
        if not success:
            return

    logger.info('Done')



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['roseingrave']

package_data = \
{'': ['*'], 'roseingrave': ['defaults/*']}

install_requires = \
['cachetools==5.2.0',
 'certifi==2022.6.15',
 'charset-normalizer==2.0.12',
 'click==8.1.3',
 'google-auth-oauthlib==0.5.2',
 'google-auth==2.9.0',
 'gspread==5.5.0',
 'idna==3.3',
 'loguru==0.6.0',
 'oauthlib==3.2.0',
 'pyasn1-modules==0.2.8',
 'pyasn1==0.4.8',
 'requests-oauthlib==1.3.1',
 'requests==2.28.0',
 'rsa==4.8',
 'six==1.16.0',
 'urllib3==1.26.9']

entry_points = \
{'console_scripts': ['roseingrave = roseingrave.__main__:cli']}

setup_kwargs = {
    'name': 'roseingrave',
    'version': '0.9.1',
    'description': 'A massively scalable document source comparator, using Google Spreadsheets API + Python.',
    'long_description': '# Roseingrave\n\nMassively scalable musical source comparator.\n\nSee the\n[documentation](https://github.com/josephlou5/roseingrave/blob/main/Documentation.md)\nfor detailed documentation.\n\n## Installation\n\nInstall the package through pip (recommended to do in a virtual environment):\n\n```bash\n$ python3 -m pip install roseingrave\n```\n\nThe package will be added as a top-level command:\n\n```bash\n$ roseingrave --help\n```\n\nCreate a folder to store all your input/ouput files. In this folder, place your\nOAuth credentials file (see\n[Credentials](https://github.com/josephlou5/roseingrave#credentials))\nand all required and optional input files.\n\nSee the\n[documentation](https://github.com/josephlou5/roseingrave/blob/main/Documentation.md)\nfor customizing filepaths and the expected input file formats.\n\n## Credentials\n\nThe package interacts with Google Sheets through the\n[`gspread` package](https://docs.gspread.org/en/latest/).\nYou can enable an OAuth Client to create, access, and edit spreadsheets with\nyour email.\n\nTo enable the OAuth Client, follow these steps:\n\n1. Go to the [Google Developers Console](https://console.cloud.google.com/).\n2. Log in with the email account you want to use with the OAuth Client. All\n   created spreadsheets will be owned by this account in Google Drive.\n3. Create a new project.\n4. Go to the [API Library](https://console.cloud.google.com/apis/library).\n5. In the search bar, search for "Google Drive API", select it, and enable it.\n6. Go back to the API library. In the search bar, search for "Google Sheets\n   API", select it, and enable it.\n7. Go to the\n   [OAuth Consent Screen](https://console.cloud.google.com/apis/credentials/consent)\n   tab.\n8. If prompted, select "External" for the User Type.\n9. On the "App Information" page, enter an app name. Select your email address\n   for the support email. Scroll down to the bottom and enter your email address\n   for the developer contact information. Click "Save and Continue".\n10. On the "Scopes" page, click "Save and Continue".\n11. On the "Test Users" page, add your email address as a user. Click "Save and\n    Continue".\n12. On the summary page, scroll to the bottom and click "Back to Dashboard".\n13. Go to the [Credentials](https://console.cloud.google.com/apis/credentials)\n    tab.\n14. At the top of the page, select "+ Create credentials" > "OAuth client ID".\n15. For the application type, select "Desktop app". Name your credentials.\n    Click "Create". Click "Ok" at the popup.\n16. In the table labeled "OAuth 2.0 Client IDs", locate the credentials you just\n    created. Click the download button at the end of the row.\n17. Rename the file to `credentials.json` and place it in the root directory of\n    where you\'ll be running the commands. (You can customize this in the\n    [settings](https://github.com/josephlou5/roseingrave/blob/main/Documentation.md#settings-optional)\n    file).\n\nIf you\'ve never authorized the app or if your authorization has expired, you\'ll\nbe given a link in the console for you to visit in order to refresh or create\nan authorization token. Go to the url, select your email, click "Continue",\nallow access to your Drive files and Sheets spreadsheets, click "Continue", copy\nthe authorization code on the final page, and paste it back into the console.\n\nOnce the authorization is successful, the `authorized_user.json` file will be\ncreated in the same directory as `"credentials"`.\n\n## Basic Usage\n\nCreate the piece definitions and volunteer definitions files as explained in the\n[documentation](https://github.com/josephlou5/roseingrave/blob/main/Documentation.md#input-files).\nIf desired, create the settings file and/or the template file. Save all the\nfiles in the proper locations as defined by the\n[default settings file](https://github.com/josephlou5/roseingrave/blob/main/src/roseingrave/defaults/roseingrave.json)\nor by your own settings file.\n\nBased on your definition files, create the volunteer spreadsheets:\n\n```bash\n$ roseingrave create_sheet\n```\n\nAfter volunteers have filled out their spreadsheets, export the data:\n\n```bash\n$ roseingrave volunteer_summary\n```\n\nExtract the data for each piece:\n\n```bash\n$ roseingrave piece_summary\n```\n\nCompile all the piece data into the summary file:\n\n```bash\n$ roseingrave compile_pieces\n```\n\nCreate the master spreadsheets with the data from the summary file:\n\n```bash\n$ roseingrave import_master\n```\n\nFill out the summary columns as appropriate, then export the master spreadsheet\ninto the summary file:\n\n```bash\n$ roseingrave export_master\n```\n\nSee the\n[commands documentation](https://github.com/josephlou5/roseingrave/blob/main/Documentation.md#commands)\nfor all commands and their arguments and options.\n',
    'author': 'Joseph Lou',
    'author_email': 'jdlou@princeton.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/josephlou5/roseingrave',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
